//
//  main.cpp
//  Freeze
//
//  Created by Charles Kerr on 9/22/20.
//

#include <iostream>
#include <filesystem>
#include <string>
#include "WSCParser.hpp"
#include <map>
#include <vector>
#include <cstdlib>

namespace fs = std::filesystem;

unsigned int heights[6] = {512,512,200,256,181,512};

void processMap(int map, std::vector<WSCObject>& objects, const std::string& uodir );
unsigned int calcBlocknumber(int& x,int& y, unsigned int height);
std::vector<WSCObject> addMulti(std::vector<WSCObject> objects, const std::string& uodir);
std::multimap<unsigned int, WSCObject> changeBlock(int map, std::vector<WSCObject>& objects);
void updateStatic(int map, std::map<unsigned int, std::vector<WSCObject>>& objects, const std::string& uodir);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int main(int argc, const char * argv[]) {
	
	if (argc < 3) {
		std::cerr <<"Requires two arguments, wsc directory and muldata directory (no spaces in filepaths)" << std::endl;
		return 1 ;
	}
	
	std::multimap<std::string, WSCObject> byWscFile;
	std::multimap<int,WSCObject> byMap ;
	std::map<std::string, std::vector<WSCObject> > objectsByFile ;

	std::string wscdir(argv[1]);
	std::string uodir(argv[2]);
	WSCParser parser ;
	auto wscpath = fs::path(wscdir) ;
	for (auto &path : fs::directory_iterator(wscpath) ) {
		if (path.path().extension() == ".wsc") {
			std::cout <<"Parsing: "<<path.path().string()<<std::endl;
			parser.parse(path.path().string());
		}
	}
	auto objects = parser.objects();
	std::cout <<"We found " <<objects.size() << " objects"<<std::endl;
	for (auto &entry : objects) {
		byMap.insert(std::make_pair(entry.loc.map,entry));
		byWscFile.insert(std::make_pair(entry.filename,entry));
	}
	{
		auto iter = byWscFile.begin();
		std::cout <<std::endl;
		while (iter != byWscFile.end()){
			std::cout <<std::endl << "We would freeze from file: "<<iter->first<<std::endl;
			auto count = byWscFile.count(iter->first) ;
			std::vector<WSCObject> obj ;
			std::string filename = iter->first ;
			for (auto i = 0 ; i< count; i++){
				std::cout <<"\t" << iter->second.description()<< " startline is " << iter->second.startline << std::endl;
				obj.push_back(iter->second);
				iter++;
			}
			objectsByFile.insert_or_assign(filename, obj);
		}
	}
	std::cout << "Should we continue (type YES for continuing)?" << std::endl;
	std::string inputstring ;
	std::cin >> inputstring ;
	if (inputstring != "YES") {
		std::cout <<"Aborting" << std::endl;
		return 1 ;
	}
	
	auto iter = objectsByFile.begin();
	char buffer[4096];
	// Rewrite the wsc files, removing the frozen items ;
	while (iter != objectsByFile.end()) {
		auto inputfile = iter->first;
		auto outputfile = inputfile + ".frozen" ;
		
		std::cout <<"Rewriting: "<<iter->first << std::endl;
		auto ent = iter->second;
		auto index = 0 ;
		auto total = ent.size() ;
		std::ifstream input(inputfile,std::ifstream::in);
		std::ofstream output(outputfile,std::ofstream::out);
		bool copy = true ;
		bool check = true ;
		int current = 0 ;
		auto present = ent[0] ;
		if (present.startline == current) {
			copy = false ;
		}
		while (!input.eof()){
			input.getline(buffer,sizeof(buffer));
			std::string line ;
			if (input.gcount()>0) {
				line = std::string(buffer) + "\n" ;
			}
			else {
				line = "\n";
			}
			std::memset(buffer,0,sizeof(buffer));
			if ((!input.eof()) && copy){
			
				output.write(line.c_str(), line.size());
			}
			current++ ;
			if (check) {
				if (current == present.startline){
					copy = false ;
				}
				else if (current > present.endline){
					copy = true ;
					index++ ;
					if (index < total) {
						present = ent[index];
						if (current == present.startline){
							copy = false ;
						}
					}
					else {
						check = false ; ;
					}
				}

			}
			
		}
		output.close();
		input.close();
		iter++;
	}

	// Now prepare the items
	// What we want to do, is get a vector of objects for each map
	auto miter = byMap.begin();
	while (miter != byMap.end()) {
		auto count = byMap.count(miter->first) ;
		auto map = miter->first ;
		std::vector<WSCObject> obj ;
		for (auto i=0 ; i < count ; i++){
			obj.push_back(miter->second);
			miter++;
		}
		// We now hae all the items for a single map
		std::cout <<"Freezing for map: " << map<<std::endl;
		processMap(map, obj, uodir);
	}
	return 0;
}

//++++++++++++++++++++++++++++++++++++++++++++
void processMap(int map, std::vector<WSCObject>& objects, const std::string& uodir ) {
	auto nobj = addMulti(objects, uodir);
	auto byBlock = changeBlock(map, nobj);
	auto iter = byBlock.begin();
	std::map<unsigned int, std::vector<WSCObject> > grouped ;
	while (iter != byBlock.end()){
		auto count = byBlock.count(iter->first) ;
		auto blocknum = iter->first ;
		std::vector<WSCObject> obj ;
		for (auto i = 0 ; i < count ; i++) {
			obj.push_back(iter->second);
			iter++ ;
		}
		grouped.insert_or_assign(blocknum, obj);
	}
	// we now have by blocknumber
	updateStatic(map, grouped, uodir);

}
//++++++++++++++++++++++++++++++++++++++++++++
void updateStatic(int map, std::map<unsigned int, std::vector<WSCObject> >& objects, const std::string& uodir) {
	auto mulfile = "statics" + std::to_string(map)+ ".mul" ;
	auto idxfile = "staidx" + std::to_string(map) + ".mul" ;
	auto mulout = mulfile + ".frozen" ;
	auto idxout = idxfile + ".frozen" ;
	auto path =std::filesystem::path(uodir);
	path /= std::filesystem::path(mulfile) ;
	std::ifstream mulinput(path.string(), std::ifstream::in | std::ifstream::binary);
	if (!mulinput.is_open()) {
		std::cerr<<"Error opening: "<<path.string() << std::endl;
		std::exit(1) ;
	}
	path = std::filesystem::path(uodir);
	path /= std::filesystem::path(idxfile) ;
	std::ifstream idxinput(path.string(), std::ifstream::in | std::ifstream::binary);
	if (!idxinput.is_open()) {
		std::cerr<<"Error opening: "<<path.string() << std::endl;
		std::exit(1) ;
	}

	path = std::filesystem::path(uodir);
	path /= std::filesystem::path(mulout) ;
	std::ofstream muloutput(path.string(), std::ofstream::out | std::ofstream::binary);

	path = std::filesystem::path(uodir);
	path /= std::filesystem::path(idxout) ;
	std::ofstream idxoutput(path.string(), std::ofstream::out | std::ofstream::binary);
	
	// so now we get to do by block number
	int lookup ;
	int length ;
	int extra ;
	unsigned int blocknum = 0 ;
	int currentoffset = 0 ;
	auto total = 0 ;
	while (!idxinput.eof()) {
		//std::cout <<"Reading header on blocknum " << blocknum<< std::endl;
		idxinput.read(reinterpret_cast<char*>(&lookup),4);
		idxinput.read(reinterpret_cast<char*>(&length),4) ;
		idxinput.read(reinterpret_cast<char*>(&extra),4) ;
		if (idxinput.gcount() != 4) {
			
			break;
		}
		auto iter = objects.find(blocknum);
		if (iter == objects.end()) {
			// That block number isn't changed
			if (lookup != -1) {
				auto buffer = new char[length];
				mulinput.read(buffer,length);
				muloutput.write(buffer,length);
				delete[] buffer ;
				idxoutput.write(reinterpret_cast<char*>(&currentoffset),4);
				idxoutput.write(reinterpret_cast<char*>(&length),4);
				idxoutput.write(reinterpret_cast<char*>(&extra),4);
				currentoffset += length ;
			}
			else {
				idxoutput.write(reinterpret_cast<char*>(&lookup),4);
				idxoutput.write(reinterpret_cast<char*>(&length),4);
				idxoutput.write(reinterpret_cast<char*>(&extra),4);

			}
		}
		else {
			// Ok, it is changed!
			//write all the statics that are there anyway
			//std::cout <<"Change on block " << blocknum << std::endl;
			if (lookup != -1) {
				auto buffer = new char[length];
				mulinput.read(buffer,length);
				muloutput.write(buffer,length);
				delete[] buffer ;
			}
			else {
				length = 0 ;
			}
			
			for (auto &entry : iter->second) {
				// For each item!
				short tileid = entry.id ;
				short color = entry.color;
				unsigned char xloc = entry.loc.x ;
				unsigned char yloc = entry.loc.y ;
				char alt = entry.loc.z ;
				total += 1 ;
				//std::cout <<"Writing count " << total << " frozen to block " << blocknum << std::endl;
				muloutput.write(reinterpret_cast<char*>(&tileid),2) ;
				muloutput.write(reinterpret_cast<char*>(&xloc),1) ;
				muloutput.write(reinterpret_cast<char*>(&yloc),1) ;
				muloutput.write(reinterpret_cast<char*>(&alt),1) ;
				muloutput.write(reinterpret_cast<char*>(&color),2) ;
				length += 7 ;
			}
			// Ok, now we write to the idx file
			idxoutput.write(reinterpret_cast<char*>(&currentoffset),4);
			idxoutput.write(reinterpret_cast<char*>(&length),4);
			idxoutput.write(reinterpret_cast<char*>(&extra),4);
			currentoffset +=length ;

		}
		blocknum++;
	}
	idxoutput.close();
	idxinput.close();
	muloutput.close();
	mulinput.close();
	
}
//++++++++++++++++++++++++++++++++++++++++++++
unsigned int calcBlocknumber(int& x,int& y, unsigned int height) {
	
	unsigned int xblock = x /8 ;
	x = x - (xblock*8);
	unsigned int yblock = y /8;
	y = y - (yblock*8) ;
	
	return (xblock * height) + yblock ;
}

//++++++++++++++++++++++++++++++++++++++++++++
std::multimap<unsigned int, WSCObject> changeBlock(int map, std::vector<WSCObject>& objects) {
	
	std::multimap<unsigned int, WSCObject> rvalue ;
	for (auto &entry : objects) {
		auto value = entry ;
		auto block = calcBlocknumber(value.loc.x, value.loc.y, heights[map]);
		
		rvalue.insert(std::make_pair(block,value));
	}
	return rvalue ;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
std::vector<WSCObject> addMulti(std::vector<WSCObject> objects, const std::string& uodir) {
	
	auto idxfile = fs::path(uodir) ;
	idxfile /= fs::path("multi.idx");
	auto mulfile = fs::path(uodir) ;
	mulfile /= fs::path("multi.mul");
	auto filesize = fs::file_size(idxfile) ;
	auto mulsize = fs::file_size(mulfile) ;
	std::ifstream idxstream(idxfile.string(),std::ifstream::in | std::ifstream::binary);
	std::ifstream mulstream(mulfile.string(),std::ifstream::in | std::ifstream::binary);
	if (!idxstream.is_open() || !mulstream.is_open()){
		std::cerr <<"Unable to open multi's at : "<<uodir<<std::endl;
		std::exit(1) ;
	}
	std::vector<WSCObject> rvalue ;
	for (auto &entry : objects) {
		if (entry.type == WSCObject::Type::item) {
			rvalue.push_back(entry) ;
		}
		else {
			auto lookup = (entry.id - 0x4000) * 12 ;
			if (filesize < (lookup+12)) {
				std::cerr <<"Error processing multi (id not in multi.idx): " << entry.description() << std::endl;
			}
			else {

				idxstream.seekg(lookup, std::ifstream::beg);
				int index ;
				int length ;
				idxstream.read(reinterpret_cast<char*>(&index), 4);
				idxstream.read(reinterpret_cast<char*>(&length), 4);
				if ((index == -1) || (static_cast<unsigned int>(index) == 0xFFFFFFFE )) {
					std::cerr <<"Error processing multi (id not in multi.idx): " << entry.description() << std::endl;
				}
				else {
					mulstream.seekg(index,std::ifstream::beg) ;
					while (length > 0) {
						short tileid  =0 ;
						short xloc =0 ;
						short yloc =0;
						short alt =0;
						
						int visible =0;
						int unknown=0 ;
						mulstream.read(reinterpret_cast<char*>(&tileid),2);
						mulstream.read(reinterpret_cast<char*>(&xloc),2);
						mulstream.read(reinterpret_cast<char*>(&yloc),2);
						mulstream.read(reinterpret_cast<char*>(&alt),2);
						mulstream.read(reinterpret_cast<char*>(&visible),4);
						length -= 12 ;
						if (mulsize >= 908592 ) {

							mulstream.read(reinterpret_cast<char*>(&unknown),4);
							length -= 4 ;
						}
						WSCObject item ;
						item.type = WSCObject::Type::item ;
						item.loc.x = entry.loc.x + xloc ;
						item.loc.y = entry.loc.y + yloc ;
						item.loc.z = entry.loc.z + alt ;
						item.loc.map = entry.loc.map ;
						item.filename = entry.filename ;
						item.id = tileid ;
						item.color = entry.color ;
						if (visible != 0) {
							std::cout <<"Multi item " << item.description()<<std::endl;
							rvalue.push_back(item );

						}
					}
					
				}
				
				
			}
		}
	}
	idxstream.close();
	mulstream.close();
	return rvalue ;
}
